# Copyright 2014-2021 The aiosmtpd Developers
# SPDX-License-Identifier: Apache-2.0

"""Test SMTP smuggling."""

from email.mime.text import MIMEText
from smtplib import SMTP, SMTP_SSL
from typing import Generator, Union

import re
import socket
import pytest
import unittest
import smtplib

from aiosmtpd.controller import Controller

from aiosmtpd.smtp import SMTP as Server
from aiosmtpd.smtp import Session as ServerSession
from aiosmtpd.smtp import Envelope


class ReceivingHandler:
    box = None

    def __init__(self):
        self.box = []

    async def handle_DATA(self, server, session, envelope):
        self.box.append(envelope)
        return '250 OK'


def new_data(self, msg):
        self.putcmd("data")

        (code, repl) = self.getreply()
        if self.debuglevel > 0:
            self._print_debug('data:', (code, repl))
        if code != 354:
            raise SMTPDataError(code, repl)
        else:
            ##### Patching input encoding so we can send raw messages
            #if isinstance(msg, str):
            #    msg = smtplib._fix_eols(msg).encode('ascii')
            #q = smtplib._quote_periods(msg)
            #if q[-2:] != smtplib.bCRLF:
            #    q = q + smtplib.bCRLF
            #q = q + b"." + smtplib.bCRLF
            q = msg
            self.send(q)
            (code, msg) = self.getreply()
            if self.debuglevel > 0:
                self._print_debug('data:', (code, msg))
            return (code, msg)


def return_unchanged(data):
    return data


def orig_data(self, msg):
    self.putcmd("data")
    (code, repl) = self.getreply()
    if self.debuglevel > 0:
        self._print_debug('data:', (code, repl))
    if code != 354:
        raise smtplib.SMTPDataError(code, repl)
    else:
        if isinstance(msg, str):
            msg = _fix_eols(msg).encode('ascii')
        q = _quote_periods(msg)
        if q[-2:] != smtplib.bCRLF:
            q = q + smtplib.bCRLF

        q = q + b"." + smtplib.bCRLF
        self.send(q)
        (code, msg) = self.getreply()
        if self.debuglevel > 0:
            self._print_debug('data:', (code, msg))
        return (code, msg)


def _fix_eols(data):
    return re.sub(r'(?:\r\n|\n|\r(?!\n))', smtplib.CRLF, data)


def _quote_periods(bindata):
    return re.sub(br'(?m)^\.', b'..', bindata)


def return_unchanged(data):
    return data


class TestSmuggling(unittest.TestCase):
    def setUp(self):
        self.handler = ReceivingHandler()
        controller = Controller(self.handler)
        controller.start()
        self.addCleanup(controller.stop)
        self.address = (controller.hostname, controller.port)

    def test_smtp_smuggling(self):
        smtplib._fix_eols = return_unchanged
        smtplib._quote_periods = return_unchanged
        smtplib.SMTP.data = new_data

        with SMTP(*self.address) as client:
            sender = "sender@example.com"
            recipients = ["rcpt1@example.com"]
            resp = client.helo("example.com")
            # S250_FQDN
            assert resp == (250, bytes(socket.getfqdn(), "utf-8"))

            # Trying SMTP smuggling with a fake \n.\r\n end-of-data sequence.
            message_data = b"""\
    From: Anne Person <anne@example.com>\r\n\
    To: Bart Person <bart@example.com>\r\n\
    Subject: A test\r\n\
    Message-ID: <ant>\r\n\
    \r\n\
    Testing\
    \n.\r\n\
    NO SMUGGLING
    \r\n.\r\n\
    """
            results = client.sendmail(sender, recipients, message_data)
            client.quit()
            smtplib._fix_eols = _fix_eols
            smtplib._quote_periods = _quote_periods
            smtplib.SMTP.data = orig_data

            assert b"NO SMUGGLING" in self.handler.box[0].content
